#pragma once

#include <string>
#include "common/BaseObject.hxx"
#include "Vector.hxx"

enum class ObjectType : uint8_t {
    None,
    Camera,
    Mesh
};

template <typename T>
class Object : public BaseObject<Object> {
public:
    // --- public data structures ---

    std::string name;
    bool visible;

    // --- constructors and deconstructors ---

    Object()
    : BaseObject(), name("unknown"), type(ObjectType::Null), visible(true),
      _origin(static_cast<const T>(0), static_cast<const T>(0), static_cast<const T>(0))
    {
    }

    Object(const std::string& n, const bool v, const Vector3<T>& o)
    : BaseObject(), name(n), visible(v), _origin(o)
    {
    }

    Object(const Object& rhs)
    : BaseObject(rhs), name(rhs.name), visible(rhs.visible), _origin(rhs._origin)
    {
    }

    Object(Object&& rhs)
    : BaseObject(rhs), name(std::move(rhs.name)), visible(std::move(rhs.visible)), _origin(std::move(rhs._origin))
    {
    }

    virtual ~Object()
    {
    }

    // --- default operators ---

    Object& operator=(const Object& rhs)
    {
        if (this != &rhs)
        {
            name = rhs.name;
            visible = rhs.visible;
            BaseObject::operator=(rhs);
        }

        return *this;
    }

    Object& operator=(Object&& rhs)
    {
        if (this != &rhs)
        {
            name = std::move(rhs.name);
            visible = std::move(rhs.visible);
            BaseObject::operator=(rhs);
        }

        return *this;
    }

    // --- public methods ---

    virtual void setOrigin(const Vector3<T>& o)
    {
        _origin = o;
    }

    virtual Vector3<T> origin() const
    {
        return _origin;
    }

    // --- interfaces ---

    virtual ObjectType type() const = 0;
    virtual void setUp() = 0;
    virtual Vector3<T> up() const = 0;

protected:
    Vector3<T> _origin;
};
