#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <GL/glew.h>
#include <GL/freeglut.h>

namespace Data
{
	enum Type {
		BSPHeader = 0,
		Faces = 11,
		VertexIndexes = 13,
		Vertices = 14,
		RGBVertices = 15,
		TextureInfo = 17,
		Textures = 18,
		TextureData = 19,
		Palettes = 23,
		PaletteEntries = 256
	};
}

struct Chunk {
	uint32_t type;
	uint32_t size;
	uint32_t count;
};

struct Vertex {
	float x, y, z;
};

struct Face {
	uint32_t first;
	uint32_t vertex_count;
	uint32_t plane_count;
	uint32_t plane_side;
	uint32_t tex_info;
	uint32_t lightmap_info;
	uint32_t lightmap_width;
	uint32_t lightmap_height;
	uint8_t lightmap_types;
};

std::vector<Face> faces;
std::vector<Vertex> vertices;
std::vector<Vertex> rgb_vertices;
std::vector<uint32_t> indexes;

void display()
{
	glClearColor(0.1, 0.1, 0.1, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glLoadIdentity();

	glTranslatef(0.0, 0.0, -200.0);
	for (Face& face : faces)
	{
		glBegin(GL_TRIANGLE_FAN);
		for (size_t j = 0; j < face.vertex_count; ++j)
		{
			Vertex& vert = vertices[indexes[face.first + j]];

			glVertex3f(vert.x / 30.0, vert.y / 30.0, vert.z / 30.0);
		}
		glEnd();
	}

	glutSwapBuffers();
}

void reshape(GLint w, GLint h)
{
	glViewport(0, 0, static_cast<GLsizei>(w), static_cast<GLsizei>(h));
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(90, static_cast<GLfloat>(w) / static_cast<GLfloat>(h), 1.0, 1000.0);
	glMatrixMode(GL_MODELVIEW);
}

int main(int argc, char **argv)
{
	std::string filename = "test.bsp";
	std::ifstream file;

	if (argc == 2)
		filename = argv[1];

	if (!std::ifstream(filename))
	{
		std::cerr << "ERROR: file not accessable" << std::endl;
		return 1;
	}

	file.open(filename);
	if (file.is_open() && file.good())
	{
		Chunk chunk;
		//size_t size = file.seekg(0, std::ios_base::end).tellg();

		file.seekg(0, std::ios_base::beg);

		while (file.read(reinterpret_cast<char *>(&chunk), sizeof (chunk)) && (chunk.type != 0xffff))
		{
			switch (chunk.type)
			{
				case Data::BSPHeader:
				{
					file.seekg(chunk.size * chunk.count, std::ios_base::cur);

					break;
				}
				case Data::Faces:
				{
					Face fac;

					for (size_t i = 0; i < chunk.count; ++i)
					{
						file.read(reinterpret_cast<char *>(&fac), chunk.size);
						faces.push_back(fac);
					}

					std::cout << "read " << chunk.count << " faces" << std::endl;

					break;
				}
				case Data::VertexIndexes:
				{
					uint32_t ind = 0;

					for (size_t i = 0; i < chunk.count; ++i)
					{
						file.read(reinterpret_cast<char *>(&ind), chunk.size);
						indexes.push_back(ind);
					}

					std::cout << "read " << chunk.count << " vertex indicies" << std::endl;

					break;
				}
				case Data::Vertices:
				{
					Vertex vert;

					for (size_t i = 0; i < chunk.count; ++i)
					{
						file.read(reinterpret_cast<char *>(&vert), chunk.size);
						vertices.push_back(vert);
					}

					std::cout << "read " << chunk.count << " vertices" << std::endl;

					break;
				}
				default:
				{
					std::cout << "dropped type " << chunk.type << " with size " << chunk.size * chunk.count << std::endl;
					file.seekg(chunk.size * chunk.count, std::ios_base::cur);

					break;
				}
			}
		}

		file.close();
	}

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(800, 600);
	glutInitWindowPosition(40, 40);
	glutCreateWindow("NC BSP Test");

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutIdleFunc(display);

	glutMainLoop();

	return 0;
}
